"""
A script that converts an Excel file into the request format.

The Excel file has to be in one of the supported formats, which examples can
be found in the data_formats directory.

To run the script on a supported Excel file, do:

    $ python excel_to_request_format.py <Excel file path> <output file path>

This prints out the request format of the excel file into `<output file path>`.

To see all the possible arguments, run:

    $ python excel_to_request_format.py --help
"""
import argparse
import json
from contextlib import suppress

import xlrd

from data_formats import request_format


class FormatUnsupportedError(Exception):
    """The Excel format is not supported."""


def convert(excel_file):
    """
    Convert the excel file into the request format.

    Args:
        excel_file: The path to the excel file

    Returns:
        The request format as a Python dict
    """
    last_exception = None
    for subclass in Converter.__subclasses__():
        converter = subclass(excel_file)

        # Assume that any errors are due to the format and nothing else
        try:
            return converter.convert()
        except Exception as exc:
            last_exception = exc

    # Found no converter for the excel file
    raise FormatUnsupportedError from last_exception


class Converter:
    """
    The SuperClass for all Excel-to-request-format converters.

    All a subclass needs to do is implement all the abstract methods and the
    superclass does the rest of the work.
    """

    def __init__(self, excel_file):
        """
        Create a new converter for the Excel file.

        Args:
            excel_file: The path to the excel file
        """
        self._excel_file = excel_file
        self._file = xlrd.open_workbook(excel_file)

    def _get_columns(self, sheet_name, start=0):
        sheet = self._file.sheet_by_name(sheet_name)

        for colx in range(start, sheet.ncols):
            yield sheet.col_values(colx)[1:]   # ipysheet didn't allow editable column names,
                                                # so added an index row to sheets and slicing it off here

    def _get_general(self):
        raise NotImplementedError

    def _get_capacities(self):
        raise NotImplementedError

    def _get_streams(self):
        raise NotImplementedError

    def _get_converters(self):
        raise NotImplementedError

    def _get_storages(self):
        raise NotImplementedError

    def _get_system_types(self):
        raise NotImplementedError

    def _get_time_series(self):
        raise NotImplementedError

    def _get_network_nodes(self):
        raise NotImplementedError

    def _get_network_links(self):
        raise NotImplementedError

    def convert(self):
        """
        Convert the file into the request format.

        Returns:
            The request format
        """
        request = {
            'version': '0.1.0',
            'general': self._get_general(),
            'capacities': self._get_capacities(),
            'streams': self._get_streams(),
            'converters': self._get_converters(),
            'storages': self._get_storages(),
            'system_types': self._get_system_types(),
            'time_series': self._get_time_series(),
            # 'network': {
            #     'nodes': self._get_network_nodes(),
            #     'links': self._get_network_links(),
            # }
        }

        return request


class _NewFormatConverter(Converter):
    def _get_columns(self, sheet_name, start=0):
        # The first column holds all the names for the rows
        return super()._get_columns(sheet_name, start=1)

    def _get_general(self):
        sheet = self._file.sheet_by_name('General')

        return {
            'interest_rate': float(sheet.cell(2, 1).value),  # since all sheets moved down one row, this has to be (2, 1)
        }

    def _get_capacities(self):
        capacities = []
        for column in self._get_columns('Capacities'):
            (name, units, item_type, options, lower_bound,
             upper_bound) = column

            if name not in ['#', '']:
                capacity = {
                    'name': name,
                    'units': units,
                    'type': item_type,
                }

                if options:
                    if item_type == 'List':
                        options = [int(x) for x in options.split(',')]

                    capacity['options'] = options

                if lower_bound != '' or upper_bound != '':
                    capacity['bounds'] = {}

                    with suppress(ValueError):
                        capacity['bounds']['lower'] = int(lower_bound)

                    with suppress(ValueError):
                        capacity['bounds']['upper'] = int(upper_bound)

                capacities.append(capacity)

        return capacities

    def _get_streams(self):
        streams = []
        for column in self._get_columns('Streams'):
            (name, importable, exportable, timeseries, price, export_price, co2, co2credit) = column

            if name not in ['#', '']:
                with suppress(ValueError):
                    name = str(name)

                with suppress(ValueError):
                    importable = int(importable)

                with suppress(ValueError):
                    exportable = int(exportable)

                with suppress(ValueError):
                    timeseries = str(timeseries)

                with suppress(ValueError):
                    price = float(price)

                with suppress(ValueError):
                    export_price = float(export_price)

                with suppress(ValueError):
                    co2 = float(co2)

                with suppress(ValueError):
                    co2credit = float(co2credit)


                stream = {
                    'name': name,
                    'importable': importable,
                    'exportable': exportable,
                    'timeseries': timeseries,
                    'price': price,
                    'export_price': export_price,
                    'co2': co2,
                    'co2_credit': co2credit
                }

                streams.append(stream)

        return streams

    def _get_converters(self):
        converters = []
        for column in self._get_columns('Converters'):
            (name, capacity, fixed_capital_cost, capital_cost,
             annual_maintenance_cost, usage_maintenance_cost, efficiency,
             lifetime, output_ratio, min_load, inputs, outputs) = column

            if name not in ['#', '']:
                converter = {
                    'name': name.strip(),
                    'efficiency': float(efficiency),
                }

                with suppress(ValueError):
                    converter['fixed_capital_cost'] = float(fixed_capital_cost)

                if capacity is not None:
                    try:
                        capacity = float(capacity)
                    except ValueError:
                        # References a capacity in the capacities
                        capacity = str(capacity)

                    converter['capacity'] = capacity

                with suppress(ValueError):
                    converter['capital_cost'] = float(capital_cost)

                with suppress(ValueError):
                    annual_maintenance_cost = float(annual_maintenance_cost)
                    converter['annual_maintenance_cost'] = annual_maintenance_cost

                with suppress(ValueError):
                    usage_maintenance_cost = float(usage_maintenance_cost)
                    converter['usage_maintenance_cost'] = usage_maintenance_cost

                with suppress(ValueError):
                    converter['lifetime'] = float(lifetime)

                with suppress(ValueError):
                    converter['output_ratio'] = float(output_ratio)

                with suppress(ValueError):
                    converter['min_load'] = float(min_load)

                if inputs:
                    converter['inputs'] = [in_.strip()
                                           for in_ in inputs.split(',')]

                if outputs:
                    converter['outputs'] = [output.strip()
                                            for output in outputs.split(',')]

                converters.append(converter)

        return converters

    def _get_storages(self):
        storages = []
        for column in self._get_columns('Storages'):
            (name, stream, capacity, cost,annual_maintenance_cost, fixed_capital_cost, lifetime, charge_efficiency,
             discharge_efficiency, decay, max_charge, max_discharge,
             min_state) = column

            if name not in ['#', '']:
                storage = {
                    'name': name,
                    'stream': stream,
                    'cost': float(cost),
                    'lifetime': float(lifetime),
                    'charge_efficiency': float(charge_efficiency),
                    'discharge_efficiency': float(discharge_efficiency),
                    'decay': float(decay),
                    'max_charge': float(max_charge),
                    'max_discharge': float(max_discharge),
                    'min_state': float(min_state),
                }

                try:
                    capacity = int(capacity)
                except ValueError:
                    # References a capacity
                    capacity = str(capacity)
                storage['capacity'] = capacity

                with suppress(ValueError):
                    annual_maintenance_cost = float(annual_maintenance_cost)
                    storage['annual_maintenance_cost'] = annual_maintenance_cost

                with suppress(ValueError):
                    fixed_capital_cost = float(fixed_capital_cost)
                    storage['fixed_capital_cost'] = fixed_capital_cost

                storages.append(storage)

        return storages

    def _get_system_types(self):
        system_types = []
        for column in self._get_columns('System types'):
            (name, *technologies) = column

            if name not in ['#', '']:
                technologies = [tech for tech in technologies
                                if tech]

                system_type = {
                    'name': name,
                    'technologies': technologies
                }
                system_types.append(system_type)

        return system_types

    def _get_time_series(self):
        time_series_list = []
        for column in self._get_columns('Time series'):
            (series_id, series_type, stream, node, units, source,
             *rest) = column

            if series_id not in ['#', '']:
                data = rest[3:]  # Remove empty lines

                time_series = {
                    'id': series_id,
                    'type': series_type,
                    'stream': stream,
                    'units': units,
                    'data': [float(d) for d in data if d is not None],
                }

                if source:
                    time_series['source'] = source

                if node:
                    time_series['node'] = int(node)

                time_series_list.append(time_series)

        return time_series_list

def parse_args():
    """Parses the command-line arguments."""
    parser = argparse.ArgumentParser(
        description='Converts an excel file into a EHub Request format.')
    parser.add_argument(
        'excel_file', help='The excel file to convert',
    )
    parser.add_argument(
        'output_file', help='The file to output the results to',
    )
    return parser.parse_args()


def main():
    """The function that runs the script."""
    args = parse_args()

    content = convert(args.excel_file)

    # Ensure the format is correct
    request_format.validate(content)

    with open(args.output_file, 'w') as file:
        file.write(json.dumps(content))

if __name__ == "__main__":
    main()



"""
A script for running as part of a CI system.

It currently only runs Pylint on any `.py` files in the directory. Pylint
checks for style and some common errors. See the `.pylintrc` file for Pylint
config options. If you have an objection with the style, make a PR to change
it.

To run the script, just do:

    $ python ci.py

in the same directory as `ci.py`.

Returns:
    A standard error code: 0 for success, non-zero for failure
"""
import os
import subprocess
from typing import Iterable


class CheckError(Exception):
    """The CI process detected an error."""


def lint(file: str) -> None:
    """
    Call pylint on a file.

    Args:
        file: The relative path of the file to `ci.py`
    """
    print(f"Running pylint on {file}")

    try:
        _ = subprocess.check_output(['pylint', file])
    except subprocess.CalledProcessError as exc:
        pylint_output = exc.stdout.decode('utf-8')

        raise CheckError(pylint_output)


def get_files_to_check(directory: str) -> Iterable[str]:
    """
    Iterate over the files to check in the directory.

    Args:
        directory: The directory to check files in

    Returns:
        An iterator over the Python files
    """
    for file in os.listdir(directory):
        # Don't show the '.'.  It's noise.
        if directory != '.':
            file = os.path.join(directory, file)

        if file.endswith('.py'):
            yield file
        elif os.path.isdir(file):
            inner_directory = file

            for inner_file in get_files_to_check(inner_directory):
                yield inner_file


def run_linting() -> int:
    """
    Run pylint on all the files.

    Returns:
        The return code
    """
    return_code = 0
    print("Running linting process...")
    for file in get_files_to_check('.'):
        try:
            lint(file)
        except CheckError as exc:
            print(exc)
            return_code = -1

    return return_code


def run_system_test():
    """
    Run all the system tests.

    Returns:
        The return code
    """
    print("Running tests...")
    result = subprocess.run(['python', '-m', 'tests.system_tests'])
    if result.returncode != 0:
        exit(result.returncode)


def run_doctests():
    """
    Run `doctest` on all the files.

    Returns:
        The return code of running all the tests.
    """
    print('Running doctests...')
    for file in get_files_to_check('.'):
        result = subprocess.run(['python', '-m', 'doctest', file])
        if result.returncode != 0:
            exit(result.returncode)

    print('Doctests PASSED')


def main() -> None:
    """The main function that runs the script."""
    return_code = run_linting()

    run_system_test()
    run_doctests()

    exit(return_code)


if __name__ == "__main__":
    main()



"""
Contains a class for a linear programming constraint.

Exports:
    >>> from pylp.constraint import Constraint

"""
from contextlib import suppress


class Constraint:
    """
    Represents a linear programming constraint.
    """

    def __init__(self, lhs, operator: str, rhs) -> None:
        """
        Create a new constraint between lhs and rhs.

        Args:
            lhs: The left-hand side of the constraint
            operator: The boolean operator as a str
            rhs: The right-hand side of the constraint
        """
        self._constraint = (lhs, operator, rhs)

    def __str__(self) -> str:
        lhs, operator, rhs = self._constraint
        return f'{lhs} {operator} {rhs}'

    @property
    def rhs(self):
        return self._constraint[2]

    @property
    def lhs(self):
        return self._constraint[0]

    @property
    def operator(self):
        return self._constraint[1]

    def construct(self):
        """Build the constraint for use in the solver."""
        lhs, operator, rhs = self._constraint

        # lhs and rhs may not be expressions
        with suppress(AttributeError):
            lhs = lhs.construct()
        with suppress(AttributeError):
            rhs = rhs.construct()

        # Lazy evaluate to avoid errors
        return {
            '<=': lambda: lhs <= rhs,
            '>=': lambda: lhs >= rhs,
            '<': lambda: lhs < rhs,
            '>': lambda: lhs > rhs,
            '==': lambda: lhs == rhs,
        }[operator]()



import argparse
import os
import numpy as np
import copy

from energy_hub.param_var import ConstantOrVar
from energy_hub import EHubModel
from energy_hub.utils import constraint
from energy_hub.input_data import InputData
from outputter import print_section, output_excel
import excel_to_request_format
import pylp
from pylp import RealVariable, BinaryVariable, IntegerVariable

import pdb

DOMAIN_TO_VARIABLE = {
    'Continuous': RealVariable,
    'Integer': IntegerVariable,
    'Binary': BinaryVariable,
}


@constraint()
def same_converter_constraint(converter, hubs):
    """
    Constraint to ensure the capacities are kept constant across all the subproblems. 
    """
    #TODO: Check that storage capacites are also kept the same
    for i in range(len(hubs)-1):
        yield hubs[i].capacities[converter] == hubs[i+1].capacities[converter]

@constraint()
def same_storage_constraint(storage, hubs):
    """
    Constraint to ensure the capacities are kept constant across all the subproblems. 
    """
    #TODO: Check that storage capacites are also kept the same
    for i in range(len(hubs)-1):
        yield hubs[i].storage_capacity[storage] == hubs[i+1].storage_capacity[storage]    


def split_hubs(excel=None, request=None, max_carbon=None, num_periods=1, len_periods=24, num_periods_in_sample_period=1, sample_period_position=0):
    """
    Splits a PyEHub into a series of smaller hubs with a given period.
    """
    if excel:
        request = excel_to_request_format.convert(excel)

    if request:
        _data = InputData(request)
    else:
        raise RuntimeError("Can't create hubs with no data.")
        
    hubs = []

    if ((num_periods*len_periods*num_periods_in_sample_period) > len(request['time_series'][0]['data'])):
        raise IndexError("Not enough data to cover all the periods.")
    
    if (num_periods_in_sample_period <= sample_period_position):
        raise IndexError("Not enough periods in sample to start at the given position.")

    for i in range(0, num_periods):
        temp_request = copy.deepcopy(request)
        for stream in temp_request['time_series']:
            stream['data'] = stream['data'][len_periods*(i + i*(num_periods_in_sample_period-1) + sample_period_position) :
                len_periods*(i+1 + i*(num_periods_in_sample_period-1) + sample_period_position)]

        hub = EHubModel(request=temp_request, max_carbon=max_carbon)
        hubs.append(hub)
    return hubs

def merge_hubs(hubs):
    """
    Compiles and combines the constraints of each subproblem into one list.
    :param hubs: List of EHubs.
    :return: The list of constraints for each hub combined with the capacities constraint to ensure the same converter capacities across all EHubs.
    """
    constraints = []
    for hub in hubs:
        hub.recompile()
        for constr in hub.constraints:
            constraints.append(constr)

    for converter in hubs[0].technologies:
        for c in same_converter_constraint(converter, hubs):
            constraints.append(c)
    
    for storage in hubs[0].storages:
        for c in same_storage_constraint(storage, hubs):
            constraints.append(c)

    return constraints

#TODO: Finish and move to outputter?
def multi_run_output(hubs, factor):
    """
    Function for collecting the correct info from the multiple EHub models and combining them into one set of results
    :param hubs: the EHub models now containing their solutions
    :param factor: If the models are samples for larger time scales
    """

    absolute_cost = hubs[0].solution_dict['investment_cost']
    for result in hubs:
        absolute_cost += result.solution_dict['maintenance_cost']*factor +  result.solution_dict['operating_cost']* factor
    return absolute_cost

def run_split_period(excel=None, request=None, output_filename=None, max_carbon=None, num_periods=1, len_periods=24, num_periods_in_sample_period=1, sample_period_position=0, solver='glpk'):
    """
    Core function for splitting a PyEHub model into smaller problems to solve together.
    :param excel: Input excel file if the hub to be split is in excel. Converted into request format before being split into subproblems.
    :param request: Input in request JSON format if the hub to be split is in JSON.
    :param output_filename: Name for file to right the output to if an output file is being used.
    :param max_carbon: Max carbon value if using a capped carbon value.
    :param num_periods: Number of sub problem EHubs to be solved together.
    :param len_periods: Number of time steps per sub problem EHub to be solved.
    :param num_periods_in_sample_period: Number of periods being grouped together to be represented by 1 sub problem EHub. Example: One week representing a whole month would be ~four periods in a sample period.
    :param sample_period_position: Which period in the grouped sample to use as the representative EHub. Example the second week of every month would be two.
    :param solver: Which MILP solver to use.
    :return: The status of pylp's solving, the list of hubs (each with their solution), and the absolute cost (cost of all the subproblems added together with the correct factor)
    """
    hubs = split_hubs(excel, request, max_carbon, num_periods, len_periods, num_periods_in_sample_period, sample_period_position)
    constraints = merge_hubs(hubs)

    #TODO: Make sure storage looping is working for each sub_hub

    objective = hubs[0].investment_cost
    for hub in hubs:
        objective += hub.operating_cost + hub.maintenance_cost


    status = pylp.solve(objective=objective, constraints=constraints, minimize=True, solver=solver)

    output = multi_run_output(hubs, num_periods_in_sample_period)

    return status, hubs, output


"""
Contains classes for linear programming variables.
"""
from numbers import Number
from typing import Union

import pulp

from pylp.expression import Expression
from pylp.constraint import Constraint

ValidOperand = Union[Expression, Number, 'Variable']


class Variable:
    """
    Represents a linear programming variable.

    This is the parent class of all variable classes that you should be using.
    ie. do NOT use this class directly.
    """
    _NAME = 0

    @classmethod
    def pulp_name(cls) -> str:
        """Return a unique name for a new variable."""
        cls._NAME += 1
        return 'x' + str(cls._NAME)

    def __init__(self, *, name: str = None, category: str = 'Continuous') -> None:
        """
        Create a new variable.

        Args:
            name: The name of the variable
            category: The type of variable to create.
        """
        if name is None:
            self._name = Variable.pulp_name()
        else:
            self._name = name
        self._category = category
        self._pulp_var = pulp.LpVariable(self._name, cat=self._category)

    def __str__(self) -> str:
        return f'({self._name} = {self.evaluate()})'

    def __add__(self, other: ValidOperand) -> Expression:
        return Expression('+', [self, other])

    def __radd__(self, other: ValidOperand) -> Expression:
        return Expression('+', [other, self])

    def __sub__(self, other: ValidOperand) -> Expression:
        return Expression('-', [self, other])

    def __rsub__(self, other: ValidOperand) -> Expression:
        return Expression('-', [other, self])

    def __neg__(self):
        return Expression('-', [0, self])

    def __mul__(self, other: ValidOperand) -> Expression:
        if isinstance(other, Variable):
            raise TypeError('Cannot multiply variables in linear programming')

        return Expression('*', [self, other])

    def __rmul__(self, other: ValidOperand) -> Expression:
        if isinstance(other, Variable):
            raise TypeError('Cannot multiply variables in linear programming')

        return Expression('*', [other, self])

    def __truediv__(self, other):
        return Expression('*', [self, 1 / other])

    def __rtruediv__(self, other):
        raise TypeError('Cannot divide by a variable in linear programming')

    def __lt__(self, other: ValidOperand) -> 'Constraint':
        return Constraint(self, '<', other)

    def __le__(self, other: ValidOperand) -> 'Constraint':
        return Constraint(self, '<=', other)

    def __gt__(self, other: ValidOperand) -> 'Constraint':
        return Constraint(self, '>', other)

    def __ge__(self, other: ValidOperand) -> 'Constraint':
        return Constraint(self, '>=', other)

    def __eq__(self, other: ValidOperand) -> 'Constraint':
        return Constraint(self, '==', other)

    def __ne__(self, other: ValidOperand) -> 'Constraint':
        raise TypeError('Cannot have != in linear programming.')

    def construct(self) -> pulp.LpVariable:
        """Build the variable for use in the solver."""
        return self._pulp_var

    def evaluate(self) -> float:
        """The value of the variable found by the solver."""
        return self._pulp_var.varValue


class RealVariable(Variable):
    """
    A variable that represents a real number.
    """

    def __init__(self, name: str = None):
        """Create a new real-valued variable."""
        super().__init__(name=name, category='Continuous')


class IntegerVariable(Variable):
    """
    A variable that represents an integer number.
    """

    def __init__(self, name: str = None):
        """Create a new inter-valued variable."""
        super().__init__(name=name, category='Integer')


class BinaryVariable(Variable):
    """
    A variable that represents a binary variable.

    ie: it only has two value: 0 or 1.
    """

    def __init__(self, name: str = None):
        """Create a new binary variable."""
        super().__init__(name=name, category='Binary')



"""
Provides functionality for handling the request format for using in the
EHubModel.
"""
from collections import defaultdict
from typing import List, Optional, Dict, TypeVar, Iterable, Union

import pandas as pd
#import logging

from data_formats.request_format import (
    Capacity, Converter, Stream, Storage, TimeSeries, Network_links
)
from energy_hub.utils import cached_property

T = TypeVar('T')


class InputData:
    """Provides convenient access to needed data to implement an energy hub
    model."""

    def __init__(self, request: dict) -> None:
        """Create a new instance.

        Args:
            request: A dictionary in request format
        """
        self._request = request

    @cached_property
    def capacities(self) -> List[Capacity]:
        """The list of capacities."""
        return [Capacity(capacity) for capacity in self._request['capacities']]

    def _get_capacity(self, name: str) -> Optional[dict]:
        """Get the capacity in the request format with this name.

        Args:
            name: The name of the capacity

        Returns:
            Either the capacity in the request format or None
        """
        capacities = self._request['capacities']

        for capacity in capacities:
            if capacity['name'] == name:
                return capacity

        return None

    @cached_property
    def storages(self) -> List[Storage]:
        """The list of storages."""
        def _make(storage):
            capacity = self._get_capacity(storage['capacity'])

            return Storage(storage, capacity)

        return [_make(storage) for storage in self._request['storages']]

    @property
    def storage_names(self) -> List[str]:
        """Return the names of the storages."""
        return [storage.name for storage in self.storages]

    @cached_property
    def converters(self) -> List[Converter]:
        """The list of converters."""
        def _make(converter):
            capacity = self._get_capacity(converter['capacity'])

            return Converter(converter, capacity)

        return [_make(converter) for converter in self._request['converters']]

    @property
    def converter_names(self) -> List[str]:
        """Return the names of the converters."""
        return [converter.name for converter in self.converters]

    @cached_property
    def time_series_list(self) -> List[TimeSeries]:
        """The list of time series."""
        return [TimeSeries(time_series)
                for time_series in self._request['time_series']]

    @cached_property
    def streams(self) -> List[Stream]:
        """The list of streams."""
        return [Stream(stream, self._request)
                for stream in self._request['streams']]

    @cached_property
    def stream_names(self) -> List[str]:
        """Return the names of the streams."""
        return [stream.name for stream in self.streams]

    @cached_property
    def output_stream_names(self):
        """The sorted list of output streams names."""
        names = (output for tech in self.converters for output in tech.outputs)

        return sorted(set(names))

    @cached_property
    def demand_stream_names(self):
        """The sorted list of demand streams names."""
        names = (demand.stream for demand in self.demands)
        return sorted(set(names))

    @cached_property
    def import_streams(self) -> List[str]:
        """The names of streams that are importable."""
        return [stream.name for  stream in self.streams
                if stream.importable]

    @cached_property
    def export_streams(self) -> List[str]:
        """The names of streams that are exportable."""
        return [stream.name for  stream in self.streams
                if stream.exportable]

    @property
    def demands(self) -> Iterable[TimeSeries]:
        """Return the TimeSeries that are demands."""
        return (demand for demand in self.time_series_list
                if demand.is_demand)

    @property
    def source(self) -> Iterable[TimeSeries]:
        """Return the TimeSeries that are sources."""
        return (source for source in self.time_series_list
                if source.is_source)

    @cached_property
    def source_stream_names(self):
        """The sorted list of source streams names."""
        names = (source.stream for source in self.source)
        return sorted(set(names))

    @cached_property
    def time_series_data(self) -> Dict[str, Dict[int, float]]:
        """The data for the availability time series as a dictionary that is indexed
        by time."""
        return {
            time_series.name: {
                t: time_series.data[t] for t in self.time
            } for time_series in self.time_series_list
        }
        #return {t: solar.data[t] for t in self.time}

    # @property
    # def availabilities(self) -> Iterable[TimeSeries]:
    #     """Return the TimeSeries that are stream availability."""
    #     return (avail for avail in self.time_series_list
    #             if avail.is_source)

    @cached_property
    def time(self) -> List[int]:
        """Return the time steps of the model."""
        # Assume all time series have data for all time steps
        example_series = list(self.demands)[0]
        time = example_series.data.keys()

        return list(time)

    @property
    def loads(self) -> Dict[str, Dict[int, float]]:
        """The data for all demands as a dictionary that is indexed by (demand
        time series ID, time)."""
        return {
            demand.stream: {
                t: demand.data[t] for t in self.time
            } for demand in self.demands
        }

    # @cached_property
    # def roof_tech(self) -> List[str]:
    #     """The names of the converters that can be put on a roof."""
    #     return [tech.name for tech in self.converters if tech.is_roof_tech]

    @property
    def c_matrix(self) -> Dict[str, Dict[str, float]]:
        """Return a dictionary-format for the C matrix.

        The format is like {converter name: {stream name: ...}, ...}
        """
        c_matrix = pd.DataFrame(0, index=self.converter_names,
                                columns=self.stream_names, dtype=float)
        for tech in self.converters:
            efficiency = tech.efficiency

            for input_ in tech.inputs:
                if input_ not in self.source_stream_names:
                    c_matrix[input_][tech.name] = -1

            for output in tech.outputs:
                output_ratio = tech.output_ratios[output]
                c_matrix[output][tech.name] = efficiency * output_ratio

        return c_matrix.T.to_dict()

    @cached_property
    def part_load(self) -> Dict[str, Dict[str, float]]:
        """Return the part load for each tech and each of its outputs."""
        part_load = defaultdict(defaultdict)  # type: Dict[str, Dict[str, float]]
        for tech in self.converters:
            for output_stream in self.output_stream_names:
                if output_stream in tech.outputs:
                    min_load = tech.min_load
                    if min_load is not None:
                        part_load[tech.name][output_stream] = min_load

        return part_load

    @property
    def converters_capacity(self) -> Dict[str, float]:
        """Return the capacities of the converters."""
        return {tech.name: tech.capacity
                for tech in self.converters}

    @property
    def stream_timeseries(self) -> Dict[str, float]:
        """Return the streams that have an availability timeseries."""
        return {stream.name: stream.timeseries
                for stream in self.streams
                if stream.timeseries}

    @property
    def part_load_techs(self) -> List[str]:
        """The names of the converters that have a part load."""
        return [tech.name for tech in self.converters if tech.has_part_load]

    @cached_property
    def linear_cost(self) -> Dict[str, float]:
        """Return the linear cost for each tech."""
        return {converter.name: converter.capital_cost
                for converter in self.converters}

    @cached_property
    def fixed_capital_costs(self) -> Dict[str, float]:
        """Return the fixed capital cost for each converter."""
        return {converter.name: converter.fixed_capital_cost
                for converter in self.converters}

    @cached_property
    def interest_rate(self) -> float:
        """The interest rate."""
        return self._request['general']['interest_rate']

    def _calculate_npv(self, lifetime: float) -> float:
        """Calculate the net present value of an asset giving its lifetime.

        Args:
            lifetime: The lifetime of the asset

        Returns:
            The net present value of the asset
        """
        r = self.interest_rate

        return 1 / (
            ((1 + r)**lifetime - 1) / (r * (1 + r)**lifetime)
        )

    @cached_property
    def tech_npv(self) -> Dict[str, float]:
        """The net present value of each converter."""
        return {tech.name: round(self._calculate_npv(tech.lifetime), 4)
                for tech in self.converters}

    @cached_property
    def variable_maintenance_cost(self) -> Dict[str, float]:
        """The variable maintenance cost of each converter."""
        return {tech.name: tech.usage_maintenance_cost
                for tech in self.converters}

    @cached_property
    def carbon_factors(self) -> Dict[str, Union[float, str]]:
        """The carbon factor of each stream."""
        carbon_factors = {}
        for stream in self.streams:
            carbon_factors[stream.name] = stream.co2

        return carbon_factors

    @cached_property
    def carbon_credits(self) -> Dict[str, Union[float, str]]:
        """The carbon credit of each stream."""
        carbon_credits = {}
        for stream in self.streams:
            carbon_credits[stream.name] = stream.co2_credit

        return carbon_credits

    @cached_property
    def fuel_price(self) -> Dict[str, Union[float, str]]:
        """Return the price of each fuel."""
        return {stream.name: stream.price
                for stream in self.streams}

    @cached_property
    def feed_in(self) -> Dict[str, Union[float, str]]:
        """The export price of each output stream."""
        return {stream.name: stream.export_price
                for stream in self.streams
                if stream.exportable}
    @property
    def storage_capacity(self):
        """The capacity of each storage."""
        return self._get_from_storages('capacity')

    @cached_property
    def storage_charge(self) -> Dict[str, float]:
        """The maximum charge of each storage."""
        return self._get_from_storages('max_charge')

    @cached_property
    def storage_discharge(self) -> Dict[str, float]:
        """The maximum discharge of each storage."""
        return self._get_from_storages('max_discharge')

    @cached_property
    def storage_loss(self) -> Dict[str, float]:
        """The decay of each storage."""
        return self._get_from_storages('decay')

    @cached_property
    def storage_ef_ch(self) -> Dict[str, float]:
        """The charging efficiency of each storage."""
        return self._get_from_storages('charge_efficiency')

    @cached_property
    def storage_ef_disch(self) -> Dict[str, float]:
        """The discharging efficiency of each storage."""
        return self._get_from_storages('discharge_efficiency')

    @cached_property
    def storage_min_soc(self) -> Dict[str, float]:
        """The minimum state of charge of each storage."""
        return self._get_from_storages('min_state')

    @cached_property
    def storage_lin_cost(self) -> Dict[str, float]:
        """The linear cost of each storage."""
        return self._get_from_storages('cost')

    @cached_property
    def storage_annual_maintenance_cost(self)->Dict[str, float]:
        """Returns annual maintenance cost of each storage"""
        return self._get_from_storages('annual_maintenance_cost')

    @cached_property
    def storage_fixed_capital_cost(self)-> Dict[str, float]:
        """Returns fixed capital cost of each storage"""
        return self._get_from_storages('fixed_capital_cost')

    def _get_from_storages(self, attribute: str) -> Dict[str, T]:
        """Return the attribute of each storage as a dictionary."""
        return {storage.name: getattr(storage, attribute)
                for storage in self.storages}

    @cached_property
    def storage_npv(self) -> Dict[str, float]:
        """The net present value of each storage."""
        return {storage.name: self._calculate_npv(storage.lifetime)
                for storage in self.storages}

    @property
    def links(self) -> List[Network_links]:
        def _make(link):
            capacity = self._get_capacity(link['capacity'])

            return Network_links(link, capacity)

        return [_make(link) for link in self._request['links']]

    @cached_property
    def links_ids(self)->List[int]:
        """Returns a list of id of all the links"""
        return [link.link_id for link in self.links]

    def _get_from_links(self, attribute: str) -> Dict[int, T]:
        return {link.link_id: getattr(link, attribute)
                for link in self.links}

    @cached_property
    def link_length(self) ->  Dict[int, float]:
        """Returns the length of each link"""
        return self._get_from_links('length')

    @cached_property
    def link_capacity(self) -> Dict[int, float]:
        """Return the capacities of the converters."""
        return {link.link_id: link.link_capacity
                for link in self.links}

    @cached_property
    def link_start(self) -> Dict[int, int]:
        """Return the id of the start node"""
        return self._get_from_links('start_id')

    @cached_property
    def link_end(self)->Dict[int, int]:
        """Return the id of the end node"""
        return self._get_from_links('end_id')


    @cached_property
    def link_thermal_loss(self) -> [int, float]:
        """Return the thermal loss of each link"""
        return self._get_from_links('total_thermal_loss')

    @cached_property
    def link_type(self) -> [int, str]:
        """Return the type of each link"""
        return self._get_from_links('link_type')

    @cached_property
    def link_reactance(self) -> [int, float]:
        """ Return the reactance of each link"""
        return self._get_from_links('link_reactance')

    @cached_property
    def fixed_network_investment_cost(self)-> float:
        return self._request['network']['fixed_network_investment_cost']

    @cached_property
    def link_proportional_cost(self) -> float:
        return self._request['network']['link_proportional_cost']



"""
This file serves as a tutorial/example of using PyLP.

To run the file from the root directory of this project, do

    python3.6 -m pylp.tutorial

This is because this tutorial is inside the package it is using.
"""
import pylp
from pylp import RealVariable, IntegerVariable


def main():
    """The main function of the tutorial.

    For this tutorial, we will be finding a selection of candy that minimizes
    our costs but we get at least 400 grams of sugar.
    """
    sugar_amount = 400  # We want to have at least this amount of sugar

    # The candy shop only has a selection of chocolate bars, lollipops, and ice
    # cream.
    sweets = ['Chocolate Bar', 'Lollipop', 'Ice Cream']

    sugar_per_sweet = {
        'Chocolate Bar': 40,  # grams per bar
        'Lollipop': 20,  # grams per item
        'Ice Cream': 0.20,  # grams per gram
    }
    cost_per_sweet = {
        'Chocolate Bar': 2.40,  # dollars per bar
        'Lollipop': 1.20,  # dollars per item
        'Ice Cream': 0.02,  # dollars per gram
    }
    amount_per_sweet = {
        'Chocolate Bar': IntegerVariable(),  # Can't have a fraction of a bar
        'Lollipop': IntegerVariable(),  # Can't have a fraction of a lollipop
        'Ice Cream': RealVariable(),  # Amount in grams
    }

    # The total cost of our selection
    total_cost = sum(cost_per_sweet[sweet] * amount_per_sweet[sweet]
                     for sweet in sweets)

    # Constrain our model so we get at least 400 grams of sugar
    sugar_constraint = sum(sugar_per_sweet[sweet] * amount_per_sweet[sweet]
                           for sweet in sweets) >= sugar_amount

    # Ensure we can't have negative amounts of anything. By defining a
    # variable, we don't constrain its values at all.
    amount_constraints = [amount_per_sweet[sweet] >= 0 for sweet in sweets]

    # Our total list of constraints
    constraints = [sugar_constraint] + amount_constraints

    # Here we solve the problem by minimizing the total cost with respect to
    # the list of constraints. For us, we will use GLPK to solve the problem.
    status = pylp.solve(objective=total_cost, constraints=constraints,
                        minimize=True, solver='glpk')

    # Print the status of the model
    print(f'Terminated in {status.time}s with a(n) {status.status} '
          f'solution/problem')

    for sweet in sweets:
        print(f'Amount of {sweet}: {amount_per_sweet[sweet].evaluate()}')

    print(f'Total Cost: {total_cost.evaluate()}')

    # You can play around with the constraints just like you would with any
    # regular Python boolean comparisons.


if __name__ == '__main__':
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyehub",
    version="1.4.1",
    author_email="revins@uvic.ca",
    description="A library used to solve an energy hub model in Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/energyincities/pyehub",
    packages=setuptools.find_packages(exclude="pytests"),
    include_package_data=True,
    install_requires=[
        'pulp',
        'contexttimer',
        'pandas',
        'numpy',
        'PyYAML',
        'xlrd',
        'jsonschema',
        'pylint',
        'openpyxl',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
)



"""
A simple library for doing mixed-integer linear programming.

Exports:
    All the variables that can be used
    >>> from pylp import (RealVariable, BinaryVariable, IntegerVariable,
    ... Variable)

    The function to solve a linear programming model and its status
    >>> from pylp import solve, Status
"""
from pylp.variable import (
    Variable, RealVariable, BinaryVariable,
    IntegerVariable,
)
from pylp.problem import solve, Status



"""
Provides functionality for handling a request format's storage.
"""
from typing import Union


class Storage:
    """A wrapper for a request format storage."""

    def __init__(self, storage_request: dict, capacity_request: dict) -> None:
        """
        Create a new wrapper.

        Args:
            storage_request: The request format storage
            capacity_request: The request format capacity corresponding to the
                storage
        """
        self._storage = storage_request
        self._capacity = capacity_request

    @property
    def name(self) -> str:
        """Return the name of the storage."""
        return self._storage['name']

    @property
    def capacity(self) -> Union[float, str]:
        """Return the capacity of the storage."""
        return self._storage['capacity']

    @property
    def stream(self) -> str:
        """The stream that this storage holds."""
        return self._storage['stream']

    @property
    def min_state(self) -> float:
        """The minimum state of charge of the storage."""
        return self._storage['min_state']

    @property
    def discharge_efficiency(self) -> float:
        """The discharge efficiency of the storage."""
        return self._storage['discharge_efficiency']

    @property
    def charge_efficiency(self) -> float:
        """The charge efficiency of the storage."""
        return self._storage['charge_efficiency']

    @property
    def lifetime(self) -> float:
        """The life time in years of the storage."""
        return self._storage['lifetime']

    @property
    def cost(self) -> float:
        """The cost of the storage."""
        return self._storage['cost']

    @property
    def max_charge(self) -> float:
        """The maximum charge of the storage."""
        return self._storage['max_charge']

    @property
    def max_discharge(self) -> float:
        """The maximum discharge of the storage."""
        return self._storage['max_discharge']

    @property
    def decay(self) -> float:
        """The decay of the storage."""
        return self._storage['decay']

    @property
    def annual_maintenance_cost(self)->float:
        """The annual maintenance cost of the storage"""
        if 'annual_maintenance_cost' in self._storage:
            return self._storage['annual_maintenance_cost']
        return 0

    @property
    def fixed_capital_cost(self)->float:
        """The fixed capital cost of the storage"""
        if 'fixed_capital_cost' in self._storage:
            return self._storage['fixed_capital_cost']

        return 0

